// AniCursor.java

// Create an animated cursor based on a Microsoft .ani file.

package ca.mb.javajeff.anicursor;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;

import java.io.*;

import java.util.*;

import javax.swing.*;

import ca.mb.javajeff.cur.*;
import ca.mb.javajeff.ico.*;
import ca.mb.javajeff.riff.*;

public class AniCursor implements Runnable
{
   private volatile boolean finished; // Animation completion flag

   private Component comp;

   private Cursor [] cursors;

   private volatile int [] durations;

   private int [] sequence;

   private volatile int index;

   private ArrayList<BufferedImage> bilist;

   private boolean isTranslucent; // true if image has an alpha channel

   private GlassPane gp;

   private int cx, cy; // Current mouse position in component relative to the
                       // glass pane

   private int hotspotx, hotspoty;

   private MouseListener ml;

   private MouseMotionListener mml;

   public AniCursor (String aniName, Component comp)
      throws BadAniException, IOException
   {
      this.comp = comp;

      if (aniName == null)
      {
          cursors = new Cursor [8];
          cursors [0] = Cursor.getPredefinedCursor (Cursor.N_RESIZE_CURSOR);
          cursors [1] = Cursor.getPredefinedCursor (Cursor.NE_RESIZE_CURSOR);
          cursors [2] = Cursor.getPredefinedCursor (Cursor.E_RESIZE_CURSOR);
          cursors [3] = Cursor.getPredefinedCursor (Cursor.SE_RESIZE_CURSOR);
          cursors [4] = Cursor.getPredefinedCursor (Cursor.S_RESIZE_CURSOR);
          cursors [5] = Cursor.getPredefinedCursor (Cursor.SW_RESIZE_CURSOR);
          cursors [6] = Cursor.getPredefinedCursor (Cursor.W_RESIZE_CURSOR);
          cursors [7] = Cursor.getPredefinedCursor (Cursor.NW_RESIZE_CURSOR);

          durations = new int [8];
          for (int i = 0; i < durations.length; i++)
               durations [i] = 75;

          sequence = new int [8];
          for (int i = 0; i < sequence.length; i++)
               sequence [i] = i;
      }
      else
      {
          RIFF riff = null;
          try
          {
              riff = new RIFF (aniName);
              if (!riff.getFormType ().equals ("ACON"))
                  throw new BadAniException ("not an animated cursor file");

              bilist = new ArrayList<BufferedImage> ();

              int x = 0;
              int y = 0;

              Chunk chunk;
              while ((chunk = riff.getChunk ()) != null)
                 if (chunk.name.equals ("anih"))
                 {
                     int size = ubyte (chunk.data [3]);
                     size <<= 8;
                     size |= ubyte (chunk.data [2]);
                     size <<= 8;
                     size |= ubyte (chunk.data [1]);
                     size <<= 8;
                     size |= ubyte (chunk.data [0]);

                     int nframes = ubyte (chunk.data [7]);
                     nframes <<= 8;
                     nframes |= ubyte (chunk.data [6]);
                     nframes <<= 8;
                     nframes |= ubyte (chunk.data [5]);
                     nframes <<= 8;
                     nframes |= ubyte (chunk.data [4]);

                     int nsteps = ubyte (chunk.data [11]);
                     nsteps <<= 8;
                     nsteps |= ubyte (chunk.data [10]);
                     nsteps <<= 8;
                     nsteps |= ubyte (chunk.data [9]);
                     nsteps <<= 8;
                     nsteps |= ubyte (chunk.data [8]);

                     int width = ubyte (chunk.data [15]);
                     width <<= 8;
                     width |= ubyte (chunk.data [14]);
                     width <<= 8;
                     width |= ubyte (chunk.data [13]);
                     width <<= 8;
                     width |= ubyte (chunk.data [12]);

                     int height = ubyte (chunk.data [19]);
                     height <<= 8;
                     height |= ubyte (chunk.data [18]);
                     height <<= 8;
                     height |= ubyte (chunk.data [17]);
                     height <<= 8;
                     height |= ubyte (chunk.data [16]);

                     int bitcount = ubyte (chunk.data [23]);
                     bitcount <<= 8;
                     bitcount |= ubyte (chunk.data [22]);
                     bitcount <<= 8;
                     bitcount |= ubyte (chunk.data [21]);
                     bitcount <<= 8;
                     bitcount |= ubyte (chunk.data [20]);

                     int nplanes = ubyte (chunk.data [27]);
                     nplanes <<= 8;
                     nplanes |= ubyte (chunk.data [26]);
                     nplanes <<= 8;
                     nplanes |= ubyte (chunk.data [25]);
                     nplanes <<= 8;
                     nplanes |= ubyte (chunk.data [24]);

                     int disprate = ubyte (chunk.data [31]);
                     disprate <<= 8;
                     disprate |= ubyte (chunk.data [30]);
                     disprate <<= 8;
                     disprate |= ubyte (chunk.data [29]);
                     disprate <<= 8;
                     disprate |= ubyte (chunk.data [28]);

                     int attr = ubyte (chunk.data [35]);
                     attr <<= 8;
                     attr |= ubyte (chunk.data [34]);
                     attr <<= 8;
                     attr |= ubyte (chunk.data [33]);
                     attr <<= 8;
                     attr |= ubyte (chunk.data [32]);

                     if ((attr & 1) == 0)
                         throw new BadAniException ("cannot deal with raw "+
                                                    "bitmap data");
                     durations = new int [nsteps];
                     for (int i = 0; i < durations.length; i++)
                          durations [i] = disprate*1000/60; // convert from
                                                            // jiffies to
                                                            // milliseconds

                     sequence = new int [nsteps];
                     for (int i = 0; i < sequence.length; i++)
                          sequence [i] = i;
                 }
                 else
                 if (chunk.name.equals ("icon"))
                 {
                     BufferedImage bi;
                     int ncolors;

                     if (chunk.data [2] == 1 && chunk.data [3] == 0)
                     {
                         Ico i;
                         try
                         {
                             i = new Ico (new ByteArrayInputStream (chunk.data));
                         }
                         catch (BadIcoResException bire)
                         {
                             throw new BadAniException (bire);
                         }
                         bi = i.getImage (0);
                         ncolors = i.getNumColors (0);
                         x = 0;
                         y = 0;
                     }
                     else
                     if (chunk.data [2] == 2 && chunk.data [3] == 0)
                     {
                         Cur c;
                         try
                         {
                             c = new Cur (new ByteArrayInputStream (chunk.data));
                         }
                         catch (BadCurResException bcre)
                         {
                             throw new BadAniException (bcre);
                         }
                         bi = c.getImage (0);
                         ncolors = c.getNumColors (0);
                         x = c.getHotspotX (0);
                         y = c.getHotspotY (0);
                     }
                     else
                        throw new BadAniException ("icon or cursor resource "+
                                                   "expected");

                     if (ncolors == 0 && !isTranslucent)
                     {
                         isTranslucent = true;
                         hotspotx = x;
                         hotspoty = y;

                         Window w = SwingUtilities.getWindowAncestor (comp);
                         if (w != null && (w instanceof JFrame))
                         {
                             gp = new GlassPane ((JFrame) w);
                             ((JFrame) w).setGlassPane (gp);
                         }
                     }

                     bilist.add (bi);
                 }
                 else
                 if (chunk.name.equals ("rate"))
                 {
                     if (durations == null)
                         throw new BadAniException ("ANI header missing");

                     for (int i = 0; i < durations.length; i++)
                     {
                          int rate = ubyte (chunk.data [4*i+3]);
                          rate <<= 8;
                          rate |= ubyte (chunk.data [4*i+2]);
                          rate <<= 8;
                          rate |= ubyte (chunk.data [4*i+1]);
                          rate <<= 8;
                          rate |= ubyte (chunk.data [4*i]);

                          durations [i] = rate*1000/60; // convert from jiffies
                                                        // to milliseconds
                     }
                 }
                 else
                 if (chunk.name.equals ("seq "))
                 {
                     if (sequence == null)
                         throw new BadAniException ("ANI header missing");

                     for (int i = 0; i < sequence.length; i++)
                     {
                          int index = ubyte (chunk.data [4*i+3]);
                          index <<= 8;
                          index |= ubyte (chunk.data [4*i+2]);
                          index <<= 8;
                          index |= ubyte (chunk.data [4*i+1]);
                          index <<= 8;
                          index |= ubyte (chunk.data [4*i]);

                          sequence [i] = index;
                     }
                 }

              cursors = new Cursor [bilist.size ()];

              Toolkit toolkit = Toolkit.getDefaultToolkit ();
              for (int i = 0; i < bilist.size (); i++)
                   cursors [i] = toolkit.
                                 createCustomCursor (bilist.get (i),
                                                     new Point (x, y),
                                                     "anicursor");
              if (gp != null)
              {
                  BufferedImage empty;
                  empty = new BufferedImage (32, 32,
                                             BufferedImage.TYPE_INT_ARGB);
                  gp.setCursor (toolkit.createCustomCursor (empty,
                                                            new Point (0, 0),
                                                            "empty"));
              }
          }
          catch (BadRIFFException briffe)
          {
             throw new BadAniException (briffe);
          }
          finally
          {
             if (riff != null)
                 riff.close ();
          }

          if (durations == null)
              throw new BadAniException ("ANI header missing");
      }
   }

   public void run ()
   {
      index = 0;

      Runnable r = new Runnable ()
                   {
                        public void run ()
                        {
                           comp.setCursor (cursors [sequence [index]]);
                           if (gp != null)
                               gp.render (bilist.get (index), cx, cy);
                        }
                   };

      while (!finished)
      {
         try
         {
             Thread.currentThread ().sleep (durations [index]);
             SwingUtilities.invokeAndWait (r);
             if (++index == durations.length)
                 index = 0;
         }
         catch (Exception ex)
         {
         }
      }

      try
      {
          r = new Runnable ()
              {
                  public void run ()
                  {
                     comp.setCursor (Cursor.
                                     getPredefinedCursor (Cursor.DEFAULT_CURSOR));
                  }
              };
          SwingUtilities.invokeAndWait (r);
      }
      catch (Exception ex)
      {
      }
   }

   public void start ()
   {
      if (isTranslucent)
      {
          ml = new MouseAdapter ()
               {
                   public void mouseEntered (MouseEvent me)
                   {
                      if (gp != null)
                          gp.setVisible (true);
                   }

                   public void mouseExited (MouseEvent me)
                   {
                      if (gp != null)
                          gp.setVisible (false);
                   }
               };
          comp.addMouseListener (ml);

          mml = new MouseMotionAdapter ()
                    {
                       public void mouseMoved (MouseEvent me)
                       {
                          if (gp != null)
                          {
                              Point pt;
                              pt = SwingUtilities.convertPoint (comp,
                                                                me.getX (),
                                                                me.getY (),
                                                                gp);
                              cx = pt.x-hotspotx;
                              cy = pt.y-hotspoty;
                          }
                       }
                    };
          comp.addMouseMotionListener (mml);
      }

      finished = false;
      new Thread (this).start ();
   }

   public void stop ()
   {
      finished = true;

      if (isTranslucent)
      {
          comp.removeMouseListener (ml);
          comp.removeMouseMotionListener (mml);
      }
   }

   private int ubyte (byte b)
   {
      return (b < 0) ? 256+b : b; // Convert byte to unsigned byte.
   }

   private class GlassPane extends JComponent
   {
      private JFrame frame;

      private BufferedImage bi;

      private int x, y;

      GlassPane (JFrame frame)
      {
         this.frame = frame;
      }

      public void paint (Graphics g)
      {
         if (bi != null)
             g.drawImage (bi, x, y, null);
      }

      void render (BufferedImage bi, int x, int y)
      {
         this.bi = bi;
         this.x = x;
         this.y = y;
         repaint ();
      }
   }
}
